import math as math

import numba
import numpy as np

SQRT_2PI = np.sqrt(2.0 * np.pi)


@numba.vectorize(['float64(float64, float64, float64)'], nopython=True)
def gaussian_pdf(x, mean, sigma):
    '''Compute the value of a Gaussian probability density function at x with
    given mean and sigma.'''
    return math.exp(-0.5 * ((x - mean) / sigma)**2) / (sigma * SQRT_2PI)


@numba.njit(parallel=True, fastmath=True, nogil=True)
def product_kde2(eval_points, samples, bandwidths):
    n_samples, n_bandwidths = samples.shape
    n_eval = eval_points.shape[0]

    assert(eval_points.shape[1] == samples.shape[1])
    assert(eval_points.shape[1] == bandwidths.shape[0])

    results = np.zeros((n_eval,))

    for eval_ind in numba.prange(n_eval):
        eval_point = eval_points[eval_ind]
        for sample_ind in range(n_samples):
            product_kde = 1.0
            sample = samples[sample_ind]
            for bandwidth_ind in range(n_bandwidths):
                bandwidth = bandwidths[bandwidth_ind]
                product_kde *= (
                    np.exp(-0.5 *
                           ((eval_point[bandwidth_ind] - sample[bandwidth_ind])
                            / bandwidth)**2) / (bandwidth * SQRT_2PI))
            results[eval_ind] += product_kde
        results[eval_ind] /= n_samples

    return results


@numba.njit(parallel=True, fastmath=True)
def product_kde(eval_points, samples, bandwidths):
    result = 0
    n_samples = len(samples)

    for sample_ind in numba.prange(n_samples):
        result += np.prod(
            gaussian_pdf(eval_points, samples[sample_ind], bandwidths)
        )

    return result / n_samples


@numba.njit(parallel=True, nogil=True)
def gauss(x, mean, sigma):
    '''Compute the value of a Gaussian probability density function at x with
    given mean and sigma.'''
    return math.exp(-0.5 * ((x - mean) / sigma)**2) / (sigma * SQRT_2PI)


@numba.njit(parallel=True, nogil=True)
def kde(test_mark, test_position, train_marks, train_position,
        place_bandwidth, mark_bandwidth, n_marks, n_position_dims,
        n_samples):
    joint_density = 0.0

    for sample_ind in range(n_samples):
        product_kde = 1.0
        for x, mean in zip(test_mark, train_marks[sample_ind]):
            product_kde *= gauss(x, mean, mark_bandwidth)

        for x, mean in zip(test_position, train_position[sample_ind]):
            product_kde *= gauss(x, mean, mark_bandwidth)
        joint_density += product_kde

    return joint_density / n_samples


@numba.njit(parallel=True, nogil=True)
def get_log_likelihood(marks, place_bin_centers, occupancy,
                       ground_intensity, train_marks, train_positions,
                       mean_rates, place_bandwidth=2.0, mark_bandwidth=20.0):
    '''

    Parameters
    ----------
    marks : ndarray, shape (n_time, n_marks, n_tetrodes, )
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    bandwidths : ndarray, shape (n_marks + n_position_dims, )
    occupancy : ndarray, shape (n_bins,)
    ground_intensity : ndarray, shape (n_tetrodes, n_bins)
    train_data : list of ndarray, shape (n_samples, n_marks +
                                            n_position_dims)
    mean_rates : ndarray, shape (n_tetrodes,)

    Returns
    -------
    log_likelihood : ndarray, shape (n_time, n_bins)

    '''
    n_time, n_marks, n_tetrodes = marks.shape
    n_bins, n_position_dims = place_bin_centers.shape
    log_likelihood = np.zeros((n_time, n_bins))

    for time_ind in numba.prange(n_time):
        for tetrode_ind in numba.prange(n_tetrodes):
            is_spike = np.any(~np.isnan(marks[time_ind, :, tetrode_ind]))
            if is_spike:
                test_mark = marks[time_ind, :, tetrode_ind]
                train_mark = train_marks[tetrode_ind]
                mean_rate = np.log(mean_rates[tetrode_ind])
                n_train = train_marks[tetrode_ind].shape[0]

                for bin_ind in numba.prange(n_bins):
                    position_bin = place_bin_centers[bin_ind, :]
                    log_likelihood[time_ind, bin_ind] += np.log(kde(
                        test_mark, position_bin, train_mark,
                        train_positions[tetrode_ind], place_bandwidth,
                        mark_bandwidth, n_marks, n_position_dims, n_train))

                    log_likelihood[time_ind, bin_ind] += mean_rate
                    log_likelihood[time_ind, bin_ind] -= np.log(
                        occupancy[bin_ind] + np.spacing(1))
                    log_likelihood[time_ind, bin_ind] -= (
                        ground_intensity[tetrode_ind][bin_ind])

            else:
                for bin_ind in range(n_bins):
                    log_likelihood[time_ind, bin_ind] -= (
                        ground_intensity[tetrode_ind][bin_ind])

    return log_likelihood



import math

import numpy as np
from numba import cuda
from numba.types import float32
from replay_trajectory_classification.bins import atleast_2d

# Precompute this constant as a float32.  Numba will inline it at compile time.
SQRT_2PI = np.float32((2 * math.pi)**0.5)
EPS = np.spacing(1)
TILE_SIZE = 128
N_BANDWIDTHS = 1


@cuda.jit(device=True)
def gaussian_pdf(x, mean, sigma):
    '''Compute the value of a Gaussian probability density function at x with
    given mean and sigma.'''
    return math.exp(-0.5 * ((x - mean) / sigma)**2) / (sigma * SQRT_2PI)


@cuda.jit
def kde_shared(eval_points, samples, bandwidths, out):
    """

    Parameters
    ----------
    eval_points : ndarray, shape (n_eval_points, n_bandwidths)
    samples : ndarray, shape (n_samples, n_bandwidths)
    bandwidths : ndarray, shape (n_bandwidths,)
    out : ndarray, shape (n_eval_points,)

    """
    n_samples = len(samples)
    n_eval_points = len(eval_points)

    n_threads_per_block = cuda.blockDim.x
    thread_index = cuda.threadIdx.x
    thread_id = thread_index + cuda.blockIdx.x * \
        n_threads_per_block  # cuda.grid(1)

    if thread_id < n_eval_points:
        # load eval point corresponding to current thread into local register
        eval_point = eval_points[thread_id]

    sum_kernel = float32(0.0)

    # for each tile of the samples
    for tile_start_ind in range(0, n_samples, TILE_SIZE):

        # load sample tile into shared memory with multiple threads
        # allocate shared memory for tile
        samples_tile = cuda.shared.array(
            shape=(TILE_SIZE, N_BANDWIDTHS), dtype=float32)
        sample_ind = tile_start_ind + thread_index

        for sub_tile_id in range(0, TILE_SIZE, n_threads_per_block):
            if ((sub_tile_id + thread_index < TILE_SIZE) &
                    (sub_tile_id + sample_ind < n_samples)):
                for bandwidth_ind in range(N_BANDWIDTHS):
                    samples_tile[sub_tile_id + thread_index, bandwidth_ind] = (
                        samples[sub_tile_id + sample_ind, bandwidth_ind])

        # make sure tile is loaded before computing the sum of kernels
        cuda.syncthreads()

        # sum the kernels for the samples in tile for the eval point
        if thread_id < n_eval_points:
            if tile_start_ind == n_samples - (n_samples % TILE_SIZE):
                n_samples_to_process = n_samples % TILE_SIZE
            else:
                n_samples_to_process = TILE_SIZE

            for sample_ind in range(n_samples_to_process):
                product_kernel = float32(1.0)
                for bandwidth_ind in range(N_BANDWIDTHS):
                    product_kernel *= gaussian_pdf(
                        eval_point[bandwidth_ind],
                        samples_tile[sample_ind, bandwidth_ind],
                        bandwidths[bandwidth_ind]
                    )
                sum_kernel += product_kernel

        # make sure computation is done before loading new tile of samples
        cuda.syncthreads()

    if thread_id < n_eval_points:
        out[thread_id] = sum_kernel / n_samples


@cuda.jit
def kde_shared1(eval_points, samples, bandwidths, out):
    """

    Parameters
    ----------
    eval_points : ndarray, shape (n_eval_points, n_bandwidths)
    samples : ndarray, shape (n_samples, n_bandwidths)
    bandwidths : ndarray, shape (n_bandwidths,)
    out : ndarray, shape (n_eval_points,)

    """
    n_samples = len(samples)
    n_eval_points = len(eval_points)

    n_threads_per_block = cuda.blockDim.x
    thread_index = cuda.threadIdx.x
    thread_id = thread_index + cuda.blockIdx.x * \
        n_threads_per_block  # cuda.grid(1)

    if thread_id < n_eval_points:
        # load eval point corresponding to current thread into local register
        eval_point = eval_points[thread_id]

    sum_kernel = float32(0.0)

    # for each tile of the samples
    for tile_start_ind in range(0, n_samples, TILE_SIZE):

        # load sample tile into shared memory with multiple threads
        # allocate shared memory for tile
        samples_tile = cuda.shared.array(shape=(TILE_SIZE, 1), dtype=float32)
        sample_ind = tile_start_ind + thread_index

        for sub_tile_id in range(0, TILE_SIZE, n_threads_per_block):
            if ((sub_tile_id + thread_index < TILE_SIZE) &
                    (sub_tile_id + sample_ind < n_samples)):
                samples_tile[sub_tile_id + thread_index,
                             0] = samples[sub_tile_id + sample_ind, 0]

        # make sure tile is loaded before computing the sum of kernels
        cuda.syncthreads()

        # sum the kernels for the samples in tile for the eval point
        if thread_id < n_eval_points:
            if tile_start_ind == n_samples - (n_samples % TILE_SIZE):
                n_samples_to_process = n_samples % TILE_SIZE
            else:
                n_samples_to_process = TILE_SIZE

            for sample_ind in range(n_samples_to_process):
                sum_kernel += gaussian_pdf(
                    eval_point[0],
                    samples_tile[sample_ind, 0],
                    bandwidths[0]
                )

        # make sure computation is done before loading new tile of samples
        cuda.syncthreads()

    if thread_id < n_eval_points:
        out[thread_id] = sum_kernel / n_samples


@cuda.jit
def kde_shared2(eval_points, samples, bandwidths, out):
    """

    Parameters
    ----------
    eval_points : ndarray, shape (n_eval_points, n_bandwidths)
    samples : ndarray, shape (n_samples, n_bandwidths)
    bandwidths : ndarray, shape (n_bandwidths,)
    out : ndarray, shape (n_eval_points,)

    """
    n_samples = len(samples)
    n_eval_points = len(eval_points)

    n_threads_per_block = cuda.blockDim.x
    thread_index = cuda.threadIdx.x
    thread_id = thread_index + cuda.blockIdx.x * \
        n_threads_per_block  # cuda.grid(1)

    if thread_id < n_eval_points:
        # load eval point corresponding to current thread into local register
        eval_point = eval_points[thread_id]

    sum_kernel = float32(0.0)

    # for each tile of the samples
    for tile_start_ind in range(0, n_samples, TILE_SIZE):

        # load sample tile into shared memory with multiple threads
        # allocate shared memory for tile
        samples_tile = cuda.shared.array(shape=(TILE_SIZE, 2), dtype=float32)
        sample_ind = tile_start_ind + thread_index

        for sub_tile_id in range(0, TILE_SIZE, n_threads_per_block):
            if ((sub_tile_id + thread_index < TILE_SIZE) &
                    (sub_tile_id + sample_ind < n_samples)):
                samples_tile[sub_tile_id + thread_index, 0] = (
                    samples[sub_tile_id + sample_ind, 0])
                samples_tile[sub_tile_id + thread_index, 1] = (
                    samples[sub_tile_id + sample_ind, 1])

        # make sure tile is loaded before computing the sum of kernels
        cuda.syncthreads()

        # sum the kernels for the samples in tile for the eval point
        if thread_id < n_eval_points:
            if tile_start_ind == n_samples - (n_samples % TILE_SIZE):
                n_samples_to_process = n_samples % TILE_SIZE
            else:
                n_samples_to_process = TILE_SIZE

            for sample_ind in range(n_samples_to_process):
                product_kernel = gaussian_pdf(
                    eval_point[0],
                    samples_tile[sample_ind, 0],
                    bandwidths[0]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[1],
                    samples_tile[sample_ind, 1],
                    bandwidths[1]
                )
                sum_kernel += product_kernel

        # make sure computation is done before loading new tile of samples
        cuda.syncthreads()

    if thread_id < n_eval_points:
        out[thread_id] = sum_kernel / n_samples


@cuda.jit
def kde_shared3(eval_points, samples, bandwidths, out):
    """

    Parameters
    ----------
    eval_points : ndarray, shape (n_eval_points, n_bandwidths)
    samples : ndarray, shape (n_samples, n_bandwidths)
    bandwidths : ndarray, shape (n_bandwidths,)
    out : ndarray, shape (n_eval_points,)

    """
    n_samples = len(samples)
    n_eval_points = len(eval_points)

    n_threads_per_block = cuda.blockDim.x
    thread_index = cuda.threadIdx.x
    thread_id = thread_index + cuda.blockIdx.x * \
        n_threads_per_block  # cuda.grid(1)

    if thread_id < n_eval_points:
        # load eval point corresponding to current thread into local register
        eval_point = eval_points[thread_id]

    sum_kernel = float32(0.0)

    # for each tile of the samples
    for tile_start_ind in range(0, n_samples, TILE_SIZE):

        # load sample tile into shared memory with multiple threads
        # allocate shared memory for tile
        samples_tile = cuda.shared.array(shape=(TILE_SIZE, 3), dtype=float32)
        sample_ind = tile_start_ind + thread_index

        for sub_tile_id in range(0, TILE_SIZE, n_threads_per_block):
            if ((sub_tile_id + thread_index < TILE_SIZE) &
                    (sub_tile_id + sample_ind < n_samples)):
                samples_tile[sub_tile_id + thread_index, 0] = (
                    samples[sub_tile_id + sample_ind, 0])
                samples_tile[sub_tile_id + thread_index, 1] = (
                    samples[sub_tile_id + sample_ind, 1])
                samples_tile[sub_tile_id + thread_index, 2] = (
                    samples[sub_tile_id + sample_ind, 2])

        # make sure tile is loaded before computing the sum of kernels
        cuda.syncthreads()

        # sum the kernels for the samples in tile for the eval point
        if thread_id < n_eval_points:
            if tile_start_ind == n_samples - (n_samples % TILE_SIZE):
                n_samples_to_process = n_samples % TILE_SIZE
            else:
                n_samples_to_process = TILE_SIZE

            for sample_ind in range(n_samples_to_process):
                product_kernel = gaussian_pdf(
                    eval_point[0],
                    samples_tile[sample_ind, 0],
                    bandwidths[0]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[1],
                    samples_tile[sample_ind, 1],
                    bandwidths[1]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[2],
                    samples_tile[sample_ind, 2],
                    bandwidths[2]
                )
                sum_kernel += product_kernel

        # make sure computation is done before loading new tile of samples
        cuda.syncthreads()

    if thread_id < n_eval_points:
        out[thread_id] = sum_kernel / n_samples


@cuda.jit
def kde_shared4(eval_points, samples, bandwidths, out):
    """

    Parameters
    ----------
    eval_points : ndarray, shape (n_eval_points, n_bandwidths)
    samples : ndarray, shape (n_samples, n_bandwidths)
    bandwidths : ndarray, shape (n_bandwidths,)
    out : ndarray, shape (n_eval_points,)

    """
    n_samples = len(samples)
    n_eval_points = len(eval_points)

    n_threads_per_block = cuda.blockDim.x
    thread_index = cuda.threadIdx.x
    thread_id = thread_index + cuda.blockIdx.x * \
        n_threads_per_block  # cuda.grid(1)

    if thread_id < n_eval_points:
        # load eval point corresponding to current thread into local register
        eval_point = eval_points[thread_id]

    sum_kernel = float32(0.0)

    # for each tile of the samples
    for tile_start_ind in range(0, n_samples, TILE_SIZE):

        # load sample tile into shared memory with multiple threads
        # allocate shared memory for tile
        samples_tile = cuda.shared.array(shape=(TILE_SIZE, 4), dtype=float32)
        sample_ind = tile_start_ind + thread_index

        for sub_tile_id in range(0, TILE_SIZE, n_threads_per_block):
            if ((sub_tile_id + thread_index < TILE_SIZE) &
                    (sub_tile_id + sample_ind < n_samples)):
                samples_tile[sub_tile_id + thread_index, 0] = (
                    samples[sub_tile_id + sample_ind, 0])
                samples_tile[sub_tile_id + thread_index, 1] = (
                    samples[sub_tile_id + sample_ind, 1])
                samples_tile[sub_tile_id + thread_index, 2] = (
                    samples[sub_tile_id + sample_ind, 2])
                samples_tile[sub_tile_id + thread_index, 3] = (
                    samples[sub_tile_id + sample_ind, 3])

        # make sure tile is loaded before computing the sum of kernels
        cuda.syncthreads()

        # sum the kernels for the samples in tile for the eval point
        if thread_id < n_eval_points:
            if tile_start_ind == n_samples - (n_samples % TILE_SIZE):
                n_samples_to_process = n_samples % TILE_SIZE
            else:
                n_samples_to_process = TILE_SIZE

            for sample_ind in range(n_samples_to_process):
                product_kernel = gaussian_pdf(
                    eval_point[0],
                    samples_tile[sample_ind, 0],
                    bandwidths[0]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[1],
                    samples_tile[sample_ind, 1],
                    bandwidths[1]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[2],
                    samples_tile[sample_ind, 2],
                    bandwidths[2]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[3],
                    samples_tile[sample_ind, 3],
                    bandwidths[3]
                )
                sum_kernel += product_kernel

        # make sure computation is done before loading new tile of samples
        cuda.syncthreads()

    if thread_id < n_eval_points:
        out[thread_id] = sum_kernel / n_samples


@cuda.jit
def kde_shared5(eval_points, samples, bandwidths, out):
    """

    Parameters
    ----------
    eval_points : ndarray, shape (n_eval_points, n_bandwidths)
    samples : ndarray, shape (n_samples, n_bandwidths)
    bandwidths : ndarray, shape (n_bandwidths,)
    out : ndarray, shape (n_eval_points,)

    """
    n_samples = len(samples)
    n_eval_points = len(eval_points)

    n_threads_per_block = cuda.blockDim.x
    thread_index = cuda.threadIdx.x
    thread_id = thread_index + cuda.blockIdx.x * \
        n_threads_per_block  # cuda.grid(1)

    if thread_id < n_eval_points:
        # load eval point corresponding to current thread into local register
        eval_point = eval_points[thread_id]

    sum_kernel = float32(0.0)

    # for each tile of the samples
    for tile_start_ind in range(0, n_samples, TILE_SIZE):

        # load sample tile into shared memory with multiple threads
        # allocate shared memory for tile
        samples_tile = cuda.shared.array(shape=(TILE_SIZE, 5), dtype=float32)
        sample_ind = tile_start_ind + thread_index

        for sub_tile_id in range(0, TILE_SIZE, n_threads_per_block):
            if ((sub_tile_id + thread_index < TILE_SIZE) &
                    (sub_tile_id + sample_ind < n_samples)):
                samples_tile[sub_tile_id + thread_index, 0] = (
                    samples[sub_tile_id + sample_ind, 0])
                samples_tile[sub_tile_id + thread_index, 1] = (
                    samples[sub_tile_id + sample_ind, 1])
                samples_tile[sub_tile_id + thread_index, 2] = (
                    samples[sub_tile_id + sample_ind, 2])
                samples_tile[sub_tile_id + thread_index, 3] = (
                    samples[sub_tile_id + sample_ind, 3])
                samples_tile[sub_tile_id + thread_index, 4] = (
                    samples[sub_tile_id + sample_ind, 4])

        # make sure tile is loaded before computing the sum of kernels
        cuda.syncthreads()

        # sum the kernels for the samples in tile for the eval point
        if thread_id < n_eval_points:
            if tile_start_ind == n_samples - (n_samples % TILE_SIZE):
                n_samples_to_process = n_samples % TILE_SIZE
            else:
                n_samples_to_process = TILE_SIZE

            for sample_ind in range(n_samples_to_process):
                product_kernel = gaussian_pdf(
                    eval_point[0],
                    samples_tile[sample_ind, 0],
                    bandwidths[0]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[1],
                    samples_tile[sample_ind, 1],
                    bandwidths[1]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[2],
                    samples_tile[sample_ind, 2],
                    bandwidths[2]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[3],
                    samples_tile[sample_ind, 3],
                    bandwidths[3]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[4],
                    samples_tile[sample_ind, 4],
                    bandwidths[4]
                )
                sum_kernel += product_kernel

        # make sure computation is done before loading new tile of samples
        cuda.syncthreads()

    if thread_id < n_eval_points:
        out[thread_id] = sum_kernel / n_samples


@cuda.jit
def kde_shared6(eval_points, samples, bandwidths, out):
    """

    Parameters
    ----------
    eval_points : ndarray, shape (n_eval_points, n_bandwidths)
    samples : ndarray, shape (n_samples, n_bandwidths)
    bandwidths : ndarray, shape (n_bandwidths,)
    out : ndarray, shape (n_eval_points,)

    """
    n_samples = len(samples)
    n_eval_points = len(eval_points)

    n_threads_per_block = cuda.blockDim.x
    thread_index = cuda.threadIdx.x
    thread_id = thread_index + cuda.blockIdx.x * \
        n_threads_per_block  # cuda.grid(1)

    if thread_id < n_eval_points:
        # load eval point corresponding to current thread into local register
        eval_point = eval_points[thread_id]

    sum_kernel = float32(0.0)

    # for each tile of the samples
    for tile_start_ind in range(0, n_samples, TILE_SIZE):

        # load sample tile into shared memory with multiple threads
        # allocate shared memory for tile
        samples_tile = cuda.shared.array(shape=(TILE_SIZE, 6), dtype=float32)
        sample_ind = tile_start_ind + thread_index

        for sub_tile_id in range(0, TILE_SIZE, n_threads_per_block):
            if ((sub_tile_id + thread_index < TILE_SIZE) &
                    (sub_tile_id + sample_ind < n_samples)):
                samples_tile[sub_tile_id + thread_index, 0] = (
                    samples[sub_tile_id + sample_ind, 0])
                samples_tile[sub_tile_id + thread_index, 1] = (
                    samples[sub_tile_id + sample_ind, 1])
                samples_tile[sub_tile_id + thread_index, 2] = (
                    samples[sub_tile_id + sample_ind, 2])
                samples_tile[sub_tile_id + thread_index, 3] = (
                    samples[sub_tile_id + sample_ind, 3])
                samples_tile[sub_tile_id + thread_index, 4] = (
                    samples[sub_tile_id + sample_ind, 4])
                samples_tile[sub_tile_id + thread_index, 5] = (
                    samples[sub_tile_id + sample_ind, 5])

        # make sure tile is loaded before computing the sum of kernels
        cuda.syncthreads()

        # sum the kernels for the samples in tile for the eval point
        if thread_id < n_eval_points:
            if tile_start_ind == n_samples - (n_samples % TILE_SIZE):
                n_samples_to_process = n_samples % TILE_SIZE
            else:
                n_samples_to_process = TILE_SIZE

            for sample_ind in range(n_samples_to_process):
                product_kernel = gaussian_pdf(
                    eval_point[0],
                    samples_tile[sample_ind, 0],
                    bandwidths[0]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[1],
                    samples_tile[sample_ind, 1],
                    bandwidths[1]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[2],
                    samples_tile[sample_ind, 2],
                    bandwidths[2]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[3],
                    samples_tile[sample_ind, 3],
                    bandwidths[3]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[4],
                    samples_tile[sample_ind, 4],
                    bandwidths[4]
                )
                product_kernel *= gaussian_pdf(
                    eval_point[5],
                    samples_tile[sample_ind, 5],
                    bandwidths[5]
                )
                sum_kernel += product_kernel

        # make sure computation is done before loading new tile of samples
        cuda.syncthreads()

    if thread_id < n_eval_points:
        out[thread_id] = sum_kernel / n_samples


KDE_FUNCS = {
    1: kde_shared1,
    2: kde_shared2,
    3: kde_shared3,
    4: kde_shared4,
    5: kde_shared5,
    6: kde_shared6,
}


def get_marks_by_place_bin_centers(marks, place_bin_centers):
    """

    Parameters
    ----------
    marks : ndarray, shape (n_spikes, n_features)
    place_bin_centers : ndarray, shape (n_position_bins, n_position_dims)

    Returns
    -------
    marks_by_place_bin_centers : ndarray, shape (n_spikes * n_position_bins,
                                                 n_features + n_position_dims)

    """
    n_spikes = marks.shape[0]
    n_place_bin_centers = place_bin_centers.shape[0]
    return np.concatenate(
        (np.tile(marks, reps=(n_place_bin_centers, 1)),
         np.repeat(place_bin_centers, n_spikes, axis=0)), axis=1)


def estimate_log_intensity(density, occupancy, mean_rate):
    '''

    Parameters
    ----------
    density : ndarray, shape (n_bins,)
    occupancy : ndarray, shape (n_bins,)
    mean_rate : float

    Returns
    -------
    intensity : ndarray, shape (n_bins,)

    '''
    return np.log(mean_rate) + np.log(density) - np.log(occupancy)


def estimate_pdf(
    decoding_marks, encoding_marks, mark_std,
    place_bin_centers, encoding_positions, position_std, stream=0
):
    '''

    Parameters
    ----------
    decoding_marks : ndarray, shape (n_decoding_spikes, n_marks)
    encoding_marks : ndarray, shape (n_encoding_spikes, n_marks)
    mark_std : float
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    encoding_positions : ndarray, shape (n_encoding_spikes, n_position_dims)
    position_std : float
    stream : numba.cuda.stream, optional

    Returns
    -------
    joint_mark_intensity : ndarray, shape (n_decoding_spikes, n_bins)

    '''
    decoding_marks = np.atleast_2d(decoding_marks)

    # Copy the arrays to the GPU
    eval_points = cuda.to_device(
        get_marks_by_place_bin_centers(decoding_marks, place_bin_centers)
        .astype(np.float32),
        stream=stream)
    encoding_samples = cuda.to_device(
        np.concatenate((encoding_marks, encoding_positions), axis=1)
        .astype(np.float32), stream=stream)

    n_marks = decoding_marks.shape[1]
    n_position_dims = place_bin_centers.shape[1]
    bandwidths = cuda.to_device(
        np.concatenate(
            ([mark_std] * n_marks,
             [position_std] * n_position_dims,
             )
        ).astype(np.float32), stream=stream)
    n_eval_points = len(eval_points)
    # Allocate memory on the GPU for the result
    pdf = cuda.device_array(
        shape=(n_eval_points,), dtype=np.float32, stream=stream)

    # Run KDE
    KDE_FUNCS[n_marks + n_position_dims].forall(n_eval_points, stream=stream)(
        eval_points, encoding_samples, bandwidths, pdf)

    return pdf


def estimate_multiunit_likelihood_gpu_shared(multiunits,
                                             encoding_marks,
                                             mark_std,
                                             place_bin_centers,
                                             encoding_positions,
                                             position_std,
                                             occupancy,
                                             mean_rates,
                                             summed_ground_process_intensity,
                                             is_track_interior=None,
                                             time_bin_size=1,
                                             n_streams=2):
    '''

    Parameters
    ----------
    multiunits : ndarray, shape (n_time, n_marks, n_electrodes)
    encoding_marks : list of ndarrays, len (n_electrodes,)
    mark_std : float
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    encoding_positions : list of ndarrays, len (n_electrodes,)
    position_std : float
    occupancy : ndarray, (n_bins,)
    mean_rates : list, len (n_electrodes,)
    summed_ground_process_intensity : ndarray, shape (n_bins,)

    Returns
    -------
    log_likelihood : ndarray, shape (n_time, n_bins)

    '''

    if is_track_interior is None:
        is_track_interior = np.ones((place_bin_centers.shape[0],),
                                    dtype=np.bool)

    n_time = multiunits.shape[0]
    log_likelihood = (-time_bin_size * summed_ground_process_intensity *
                      np.ones((n_time, 1)))
    n_electrodes = multiunits.shape[-1]
    multiunits = np.moveaxis(multiunits, -1, 0)
    streams = [cuda.stream() for _ in range(min(n_streams, n_electrodes))]
    pdfs = []
    is_spikes = []

    for elec_ind, (multiunit, enc_marks, enc_pos) in enumerate(zip(
            multiunits, encoding_marks, encoding_positions)):
        nan_multiunit = np.isnan(multiunit)
        is_spike = np.any(~nan_multiunit, axis=1)
        nan_mark_dims = np.all(nan_multiunit, axis=0)

        is_spikes.append(is_spike)
        n_spikes = is_spike.sum()
        if n_spikes > 0:
            pdfs.append(estimate_pdf(
                multiunit[np.ix_(is_spike, ~nan_mark_dims)],
                enc_marks,
                mark_std,
                place_bin_centers[is_track_interior],
                enc_pos,
                position_std,
                stream=streams[elec_ind % n_streams]
            ))
        else:
            pdfs.append([])

    n_interior_place_bins = is_track_interior.sum()
    for elec_ind, (pdf, mean_rate, is_spike) in enumerate(
            zip(pdfs, mean_rates, is_spikes)):
        n_spikes = is_spike.sum()
        if n_spikes > 0:
            # Copy results from GPU to CPU and
            # reshape to (n_decoding_spikes, n_interior_place_bins)
            pdf = (pdf
                   .copy_to_host(stream=streams[elec_ind % n_streams])
                   .reshape((n_spikes, n_interior_place_bins), order='F'))
            log_intensity = estimate_log_intensity(
                pdf,
                occupancy[is_track_interior],
                mean_rate)
            is_inf = np.all(np.isneginf(log_intensity), axis=1)
            log_intensity[is_inf] = np.spacing(1)
            log_likelihood[np.ix_(
                is_spike, is_track_interior)] += log_intensity

    log_likelihood[:, ~is_track_interior] = np.nan

    return log_likelihood


def estimate_position_density(place_bin_centers, positions, position_std):
    # Copy the arrays to the device
    eval_points = cuda.to_device(place_bin_centers.astype(np.float32))
    samples = cuda.to_device(positions.astype(np.float32))
    n_position_dims = positions.shape[1]
    bandwidths = cuda.to_device(
        np.asarray([position_std] * n_position_dims).astype(np.float32))

    # Allocate memory on the device for the result
    n_eval_points = len(eval_points)
    out = cuda.device_array((n_eval_points,), dtype=np.float32)

    KDE_FUNCS[n_position_dims].forall(n_eval_points)(
        eval_points, samples, bandwidths, out)

    return out.copy_to_host()


def estimate_intensity(density, occupancy, mean_rate):
    '''

    Parameters
    ----------
    density : ndarray, shape (n_bins,)
    occupancy : ndarray, shape (n_bins,)
    mean_rate : float

    Returns
    -------
    intensity : ndarray, shape (n_bins,)

    '''
    return np.exp(estimate_log_intensity(density, occupancy, mean_rate))


def fit_multiunit_likelihood_gpu_shared(position,
                                        multiunits,
                                        place_bin_centers,
                                        mark_std,
                                        position_std,
                                        is_track_interior=None,
                                        **kwargs):
    '''

    Parameters
    ----------
    position : ndarray, shape (n_time, n_position_dims)
    multiunits : ndarray, shape (n_time, n_marks, n_electrodes)
    place_bin_centers : ndarray, shape ( n_bins, n_position_dims)
    mark_std : float
    position_std : float
    is_track_interior : None or ndarray, shape (n_bins,)

    Returns
    -------
    encoding_model : dict

    '''
    if is_track_interior is None:
        is_track_interior = np.ones((place_bin_centers.shape[0],),
                                    dtype=np.bool)
    position = atleast_2d(position)
    place_bin_centers = atleast_2d(place_bin_centers)

    not_nan_position = np.all(~np.isnan(position), axis=1)

    occupancy = np.zeros((place_bin_centers.shape[0],), dtype=np.float32)
    occupancy[is_track_interior] = estimate_position_density(
        place_bin_centers[is_track_interior],
        position[not_nan_position],
        position_std)

    mean_rates = []
    ground_process_intensities = []
    encoding_marks = []
    encoding_positions = []

    for multiunit in np.moveaxis(multiunits, -1, 0):

        # ground process intensity
        nan_multiunit = np.isnan(multiunit)
        is_spike = np.any(~nan_multiunit, axis=1)
        nan_mark_dims = np.all(nan_multiunit, axis=0)

        mean_rates.append(is_spike.mean())
        marginal_density = np.zeros(
            (place_bin_centers.shape[0],), dtype=np.float32)

        if is_spike.sum() > 0:
            marginal_density[is_track_interior] = estimate_position_density(
                place_bin_centers[is_track_interior],
                position[is_spike & not_nan_position], position_std)

        ground_process_intensities.append(
            estimate_intensity(marginal_density, occupancy, mean_rates[-1])
            + EPS)

        encoding_marks.append(
            multiunit[np.ix_(is_spike & not_nan_position, ~nan_mark_dims)
                      ].astype(np.float32))
        encoding_positions.append(position[is_spike & not_nan_position])

    summed_ground_process_intensity = np.sum(
        np.stack(ground_process_intensities, axis=0), axis=0, keepdims=True)

    return {
        'encoding_marks': encoding_marks,
        'encoding_positions': encoding_positions,
        'summed_ground_process_intensity': summed_ground_process_intensity,
        'occupancy': occupancy,
        'mean_rates': mean_rates,
        'mark_std': mark_std,
        'position_std': position_std,
        **kwargs,
    }



import math as math

import dask
import dask.array as da
import numba
import numpy as np
from replay_trajectory_classification.bins import atleast_2d

SQRT_2PI = np.sqrt(2.0 * np.pi)


@numba.vectorize(['float64(float64, float64, float64)'], nopython=True,
                 cache=True)
def gaussian_pdf(x, mean, sigma):
    '''Compute the value of a Gaussian probability density function at x with
    given mean and sigma.'''
    return math.exp(-0.5 * ((x - mean) / sigma)**2) / (sigma * SQRT_2PI)


def estimate_position_distance(place_bin_centers, positions, position_std):
    return np.prod(
        gaussian_pdf(
            np.expand_dims(place_bin_centers, axis=0),
            np.expand_dims(positions, axis=1),
            position_std),
        axis=-1
    )


def estimate_position_density(place_bin_centers, positions, position_std):
    return np.mean(estimate_position_distance(
        place_bin_centers, positions, position_std), axis=0)


def estimate_log_intensity(density, occupancy, mean_rate):
    return np.log(mean_rate) + np.log(density) - np.log(occupancy)


def estimate_intensity(density, occupancy, mean_rate):
    '''

    Parameters
    ----------
    density : ndarray, shape (n_bins,)
    occupancy : ndarray, shape (n_bins,)
    mean_rate : float

    Returns
    -------
    intensity : ndarray, shape (n_bins,)

    '''
    return np.exp(estimate_log_intensity(density, occupancy, mean_rate))


def normal_pdf_integer_lookup(x, mean, std=20, max_value=3000):
    """Fast density evaluation for integers by precomputing a hash table of
    values.

    Parameters
    ----------
    x : int
    mean : int
    std : float
    max_value : int

    Returns
    -------
    probability_density : int

    """
    normal_density = gaussian_pdf(np.arange(-max_value, max_value), 0, std)

    return normal_density[(x - mean) + max_value]


def estimate_log_joint_mark_intensity(decoding_marks,
                                      encoding_marks,
                                      mark_std,
                                      place_bin_centers,
                                      encoding_positions,
                                      position_std,
                                      occupancy,
                                      mean_rate,
                                      max_mark_value=6000,
                                      set_diag_zero=False,
                                      position_distance=None):
    """

    Parameters
    ----------
    decoding_marks : ndarray, shape (n_decoding_spikes, n_features)
    encoding_marks : ndarray, shape (n_encoding_spikes, n_features)
    mark_std : float or ndarray, shape (n_features,)
    place_bin_centers : ndarray, shape (n_position_bins, n_position_dims)
    encoding_positions : ndarray, shape (n_decoding_spikes, n_position_dims)
    position_std : float or ndarray, shape (n_position_dims,)
    occupancy : ndarray, shape (n_position_bins,)
    mean_rate : float
    is_track_interior : None or ndarray, shape (n_position_bins,)
    max_mark_value : int
    set_diag_zero : bool

    Returns
    -------
    log_joint_mark_intensity : ndarray, shape (n_decoding_spikes, n_position_bins)

    """
    # mark_distance: ndarray, shape (n_decoding_spikes, n_encoding_spikes)
    mark_distance = np.prod(
        normal_pdf_integer_lookup(
            np.expand_dims(decoding_marks, axis=1),
            np.expand_dims(encoding_marks, axis=0),
            std=mark_std,
            max_value=max_mark_value
        ),
        axis=-1
    )

    if set_diag_zero:
        diag_ind = np.diag_indices_from(mark_distance)
        mark_distance[diag_ind] = 0.0

    n_encoding_spikes = encoding_marks.shape[0]

    if position_distance is None:
        position_distance = estimate_position_distance(
            place_bin_centers, encoding_positions, position_std)

    return estimate_log_intensity(
        mark_distance @ position_distance / n_encoding_spikes,
        occupancy,
        mean_rate)


def fit_multiunit_likelihood_integer_pass_position(position,
                                                   multiunits,
                                                   place_bin_centers,
                                                   mark_std,
                                                   position_std,
                                                   is_track_interior=None,
                                                   **kwargs):
    '''

    Parameters
    ----------
    position : ndarray, shape (n_time, n_position_dims)
    multiunits : ndarray, shape (n_time, n_marks, n_electrodes)
    place_bin_centers : ndarray, shape ( n_bins, n_position_dims)
    model : sklearn model
    model_kwargs : dict
    occupancy_model : sklearn model
    occupancy_kwargs : dict
    is_track_interior : None or ndarray, shape (n_bins,)

    Returns
    -------
    joint_pdf_models : list of sklearn models, shape (n_electrodes,)
    ground_process_intensities : list of ndarray, shape (n_electrodes,)
    occupancy : ndarray, (n_bins, n_position_dims)
    mean_rates : ndarray, (n_electrodes,)

    '''
    if is_track_interior is None:
        is_track_interior = np.ones((place_bin_centers.shape[0],),
                                    dtype=np.bool)
    position = atleast_2d(position)
    place_bin_centers = atleast_2d(place_bin_centers)

    not_nan_position = np.all(~np.isnan(position), axis=1)

    occupancy = np.zeros((place_bin_centers.shape[0],))
    occupancy[is_track_interior] = estimate_position_density(
        place_bin_centers[is_track_interior],
        position[not_nan_position],
        position_std)

    mean_rates = []
    ground_process_intensities = []
    encoding_marks = []
    encoding_positions = []
    position_at_spike_distances = []

    for multiunit in np.moveaxis(multiunits, -1, 0):

        # ground process intensity
        is_spike = np.any(~np.isnan(multiunit), axis=1)
        mean_rates.append(is_spike.mean())
        marginal_density = np.zeros((place_bin_centers.shape[0],))
        position_at_spike_distance = estimate_position_distance(
            place_bin_centers[is_track_interior],
            position[is_spike & not_nan_position],
            position_std)

        if is_spike.sum() > 0:
            marginal_density[is_track_interior] = np.mean(
                position_at_spike_distance, axis=0)

        position_at_spike_distances.append(position_at_spike_distance)
        ground_process_intensities.append(
            estimate_intensity(marginal_density, occupancy, mean_rates[-1])
            + np.spacing(1))

        encoding_marks.append(
            multiunit[is_spike & not_nan_position].astype(int))
        encoding_positions.append(position[is_spike & not_nan_position])

    summed_ground_process_intensity = np.sum(
        np.stack(ground_process_intensities, axis=0), axis=0, keepdims=True)

    return {
        'encoding_marks': encoding_marks,
        'encoding_positions': encoding_positions,
        'summed_ground_process_intensity': summed_ground_process_intensity,
        'position_at_spike_distances': position_at_spike_distances,
        'occupancy': occupancy,
        'mean_rates': mean_rates,
        'mark_std': mark_std,
        'position_std': position_std,
        **kwargs,
    }


def estimate_multiunit_likelihood_integer_pass_position(multiunits,
                                                        encoding_marks,
                                                        mark_std,
                                                        place_bin_centers,
                                                        encoding_positions,
                                                        position_std,
                                                        position_at_spike_distances,
                                                        occupancy,
                                                        mean_rates,
                                                        summed_ground_process_intensity,
                                                        max_mark_value=6000,
                                                        set_diag_zero=False,
                                                        is_track_interior=None,
                                                        time_bin_size=1,
                                                        chunks=None):
    '''

    Parameters
    ----------
    multiunits : ndarray, shape (n_time, n_marks, n_electrodes)
    place_bin_centers : ndarray, (n_bins, n_position_dims)
    joint_pdf_models : list of sklearn models, shape (n_electrodes,)
    ground_process_intensities : list of ndarray, shape (n_electrodes,)
    occupancy : ndarray, (n_bins, n_position_dims)
    mean_rates : ndarray, (n_electrodes,)

    Returns
    -------
    log_likelihood : (n_time, n_bins)

    '''

    if is_track_interior is None:
        is_track_interior = np.ones((place_bin_centers.shape[0],),
                                    dtype=np.bool)

    n_time = multiunits.shape[0]
    log_likelihood = (-time_bin_size * summed_ground_process_intensity *
                      np.ones((n_time, 1)))

    multiunits = np.moveaxis(multiunits, -1, 0)
    log_joint_mark_intensities = []

    for multiunit, enc_marks, enc_pos, mean_rate, pos_at_spike_dist in zip(
            multiunits, encoding_marks, encoding_positions, mean_rates,
            position_at_spike_distances):
        is_spike = np.any(~np.isnan(multiunit), axis=1)
        decoding_marks = da.from_array(
            multiunit[is_spike].astype(np.int))
        log_joint_mark_intensities.append(
            decoding_marks.map_blocks(
                estimate_log_joint_mark_intensity,
                enc_marks,
                mark_std,
                place_bin_centers[is_track_interior],
                enc_pos,
                position_std,
                occupancy[is_track_interior],
                mean_rate,
                position_distance=da.from_array(pos_at_spike_dist),
                max_mark_value=max_mark_value,
                set_diag_zero=set_diag_zero,
                chunks=chunks
            ))

    for log_joint_mark_intensity, multiunit in zip(
            dask.compute(*log_joint_mark_intensities), multiunits):
        is_spike = np.any(~np.isnan(multiunit), axis=1)
        log_likelihood[np.ix_(is_spike, is_track_interior)] += (
            log_joint_mark_intensity + np.spacing(1))

    log_likelihood[:, ~is_track_interior] = np.nan

    return log_likelihood



import math

import numpy as np
from numba import cuda
from numba.types import float32
from replay_trajectory_classification.bins import atleast_2d

# Precompute this constant as a float32.  Numba will inline it at compile time.
SQRT_2PI = np.float32((2 * math.pi)**0.5)
EPS = np.finfo(np.float32).eps


@cuda.jit(device=True)
def gaussian_pdf(x, mean, sigma):
    '''Compute the value of a Gaussian probability density function at x with
    given mean and sigma.'''
    return math.exp(-0.5 * ((x - mean) / sigma)**2) / (sigma * SQRT_2PI)

# cuda.jit('(float32[:, :], float32[:, :], float32[:], float32[:])')


@cuda.jit()
def kde1(eval_points, samples, bandwidths, out):
    """

    Parameters
    ----------
    eval_points : ndarray, shape (n_eval_points, n_bandwidths)
    samples : ndarray, shape (n_samples, n_bandwidths)
    bandwidths : ndarray, shape (n_bandwidths,)
    out : ndarray, shape (n_eval_points,)

    """
    n_bandwidths = len(bandwidths)
    n_samples = len(samples)
    n_eval_points = len(eval_points)

    for thread_id in range(cuda.grid(1), n_eval_points, cuda.gridsize(1)):
        eval_point = eval_points[thread_id]
        sum_kernel = float32(0.0)
        for sample_ind in range(n_samples):
            product_kernel = float32(1.0)
            for bandwidth_ind in range(n_bandwidths):
                product_kernel *= gaussian_pdf(
                    eval_point[bandwidth_ind],
                    samples[sample_ind, bandwidth_ind],
                    bandwidths[bandwidth_ind]
                )
            sum_kernel += product_kernel

        out[thread_id] = sum_kernel / n_samples


def get_marks_by_place_bin_centers(marks, place_bin_centers):
    """

    Parameters
    ----------
    marks : ndarray, shape (n_spikes, n_features)
    place_bin_centers : ndarray, shape (n_position_bins, n_position_dims)

    Returns
    -------
    marks_by_place_bin_centers : ndarray, shape (n_spikes * n_position_bins,
                                                 n_features + n_position_dims)

    """
    n_spikes = marks.shape[0]
    n_place_bin_centers = place_bin_centers.shape[0]
    return np.concatenate(
        (np.tile(marks, reps=(n_place_bin_centers, 1)),
         np.repeat(place_bin_centers, n_spikes, axis=0)), axis=1)


def estimate_log_intensity(density, occupancy, mean_rate):
    '''

    Parameters
    ----------
    density : ndarray, shape (n_bins,)
    occupancy : ndarray, shape (n_bins,)
    mean_rate : float

    Returns
    -------
    intensity : ndarray, shape (n_bins,)

    '''
    return np.log(mean_rate) + np.log(density) - np.log(occupancy)


def pin_arrays(
    decoding_marks, encoding_marks, place_bin_centers, encoding_positions,
    bandwidths, stream=0
):
    '''

    Parameters
    ----------
    decoding_marks : ndarray, shape (n_decoding_spikes, n_marks)
    encoding_marks : ndarray, shape (n_encoding_spikes, n_marks)
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    encoding_positions : ndarray, shape (n_encoding_spikes, n_position_dims)
    bandwidths : numba.cuda.const
    stream : numba.cuda.stream, optional

    Returns
    -------
    joint_mark_intensity : ndarray, shape (n_decoding_spikes, n_bins)

    '''
    decoding_marks = np.atleast_2d(decoding_marks)
    eval_points = (get_marks_by_place_bin_centers(decoding_marks, place_bin_centers)
                   .astype(np.float32))
    encoding_samples = (np.concatenate((encoding_marks, encoding_positions), axis=1)
                        .astype(np.float32))

    n_decoding_spikes, n_marks = decoding_marks.shape
    n_bins, n_position_dims = place_bin_centers.shape
    n_eval_points = len(eval_points)

    pdf = np.empty((n_eval_points,), dtype=np.float32)

    with cuda.pinned(eval_points, encoding_samples, bandwidths, pdf):
        # Copy the arrays to the GPU
        d_eval_points = cuda.to_device(eval_points, stream=stream)
        d_encoding_samples = cuda.to_device(encoding_samples, stream=stream)
        d_bandwidths = cuda.to_device(bandwidths, stream=stream)

        # Allocate memory on the GPU for the result
        d_pdf = cuda.device_array_like(pdf, stream=stream)

    return (d_eval_points, d_encoding_samples, d_bandwidths, d_pdf,
            n_decoding_spikes, pdf)


def estimate_multiunit_likelihood_gpu_pinned_arrays(multiunits,
                                                    encoding_marks,
                                                    mark_std,
                                                    place_bin_centers,
                                                    encoding_positions,
                                                    position_std,
                                                    occupancy,
                                                    mean_rates,
                                                    summed_ground_process_intensity,
                                                    is_track_interior=None,
                                                    time_bin_size=1,
                                                    n_streams=16):
    '''

    Parameters
    ----------
    multiunits : ndarray, shape (n_time, n_marks, n_electrodes)
    encoding_marks : list of ndarrays, len (n_electrodes,)
    mark_std : float
    place_bin_centers : ndarray, shape (n_bins, n_position_dims)
    encoding_positions : list of ndarrays, len (n_electrodes,)
    position_std : float
    occupancy : ndarray, (n_bins,)
    mean_rates : list, len (n_electrodes,)
    summed_ground_process_intensity : ndarray, shape (n_bins,)

    Returns
    -------
    log_likelihood : ndarray, shape (n_time, n_bins)

    '''

    if is_track_interior is None:
        is_track_interior = np.ones((place_bin_centers.shape[0],),
                                    dtype=np.bool)

    n_time, n_marks, n_electrodes = multiunits.shape
    log_likelihood = (-time_bin_size * summed_ground_process_intensity *
                      np.ones((n_time, 1)))
    streams = [cuda.stream() for _ in range(min(n_streams, n_electrodes))]

    n_position_dims = place_bin_centers.shape[1]
    bandwidths = (np.concatenate(
        ([mark_std] * n_marks,
         [position_std] * n_position_dims,
         )
    ).astype(np.float32))

    device_arrays = []
    with cuda.defer_cleanup():
        for elec_ind, (multiunit, enc_marks, enc_pos, mean_rate) in enumerate(zip(
                np.moveaxis(multiunits, -1, 0), encoding_marks, encoding_positions, mean_rates)):
            nan_multiunit = np.isnan(multiunit)
            is_spike = np.any(~nan_multiunit, axis=1)
            nan_mark_dims = np.all(nan_multiunit, axis=0)

            if is_spike.sum() > 0:
                device_arrays.append(pin_arrays(
                    multiunit[np.ix_(is_spike, ~nan_mark_dims)],
                    enc_marks,
                    place_bin_centers[is_track_interior],
                    enc_pos,
                    bandwidths,
                    stream=streams[elec_ind % n_streams]
                ))
            else:
                device_arrays.append([[], [], [], [], [], []])

        # Run KDE
        for elec_ind, (d_eval_points, d_encoding_samples, d_bandwidths, d_pdf,
                       _, _) in enumerate(device_arrays):
            if len(d_eval_points) > 0:
                stream = streams[elec_ind % n_streams]
                n_eval_points = d_eval_points.shape[0]

                kde1.forall(n_eval_points, stream=stream)(
                    d_eval_points, d_encoding_samples, d_bandwidths, d_pdf)

        for elec_ind, (_, _, _, d_pdf, _, pdf) in enumerate(
                device_arrays):
            if len(d_pdf) > 0:
                stream = streams[elec_ind % n_streams]
                d_pdf.copy_to_host(pdf, stream=stream)

    n_bins = np.sum(is_track_interior)

    for elec_ind, ((_, _, _, _, n_decoding_spikes, pdf), multiunit, mean_rate) in enumerate(
            zip(device_arrays, np.moveaxis(multiunits, -1, 0), mean_rates)):
        is_spike = np.any(~np.isnan(multiunit), axis=1)
        if is_spike.sum() > 0:
            log_intensity = (
                estimate_log_intensity(
                    pdf.reshape((n_decoding_spikes, n_bins), order='F') + EPS,
                    occupancy[is_track_interior] + EPS,
                    mean_rate))
            is_inf = np.all(np.isneginf(log_intensity), axis=1)
            log_intensity[is_inf] = np.spacing(1)
            log_likelihood[np.ix_(
                is_spike, is_track_interior)] += log_intensity

    log_likelihood[:, ~is_track_interior] = np.nan

    return log_likelihood


def estimate_position_density(place_bin_centers, positions, position_std):
    # Copy the arrays to the device
    eval_points = cuda.to_device(place_bin_centers.astype(np.float32))
    samples = cuda.to_device(positions.astype(np.float32))
    n_position_dims = positions.shape[1]
    bandwidths = cuda.to_device(
        np.asarray([position_std] * n_position_dims).astype(np.float32))

    # Allocate memory on the device for the result
    n_eval_points = len(eval_points)
    out = cuda.device_array((n_eval_points,), dtype=np.float32)

    kde1.forall(n_eval_points)(
        eval_points, samples, bandwidths, out)

    return out.copy_to_host()


def estimate_intensity(density, occupancy, mean_rate):
    '''

    Parameters
    ----------
    density : ndarray, shape (n_bins,)
    occupancy : ndarray, shape (n_bins,)
    mean_rate : float

    Returns
    -------
    intensity : ndarray, shape (n_bins,)

    '''
    return np.exp(estimate_log_intensity(density, occupancy, mean_rate))


def fit_multiunit_likelihood_gpu_pinned_arrays(position,
                                               multiunits,
                                               place_bin_centers,
                                               mark_std,
                                               position_std,
                                               is_track_interior=None,
                                               **kwargs):
    '''

    Parameters
    ----------
    position : ndarray, shape (n_time, n_position_dims)
    multiunits : ndarray, shape (n_time, n_marks, n_electrodes)
    place_bin_centers : ndarray, shape ( n_bins, n_position_dims)
    mark_std : float
    position_std : float
    is_track_interior : None or ndarray, shape (n_bins,)

    Returns
    -------
    encoding_model : dict

    '''
    if is_track_interior is None:
        is_track_interior = np.ones((place_bin_centers.shape[0],),
                                    dtype=np.bool)
    position = atleast_2d(position)
    place_bin_centers = atleast_2d(place_bin_centers)

    not_nan_position = np.all(~np.isnan(position), axis=1)

    occupancy = np.zeros((place_bin_centers.shape[0],), dtype=np.float32)
    occupancy[is_track_interior] = estimate_position_density(
        place_bin_centers[is_track_interior],
        position[not_nan_position],
        position_std)

    mean_rates = []
    ground_process_intensities = []
    encoding_marks = []
    encoding_positions = []

    for multiunit in np.moveaxis(multiunits, -1, 0):

        # ground process intensity
        nan_multiunit = np.isnan(multiunit)
        is_spike = np.any(~nan_multiunit, axis=1)
        nan_mark_dims = np.all(nan_multiunit, axis=0)

        mean_rates.append(is_spike.mean())
        marginal_density = np.zeros(
            (place_bin_centers.shape[0],), dtype=np.float32)

        if is_spike.sum() > 0:
            marginal_density[is_track_interior] = estimate_position_density(
                place_bin_centers[is_track_interior],
                position[is_spike & not_nan_position], position_std)

        ground_process_intensities.append(
            estimate_intensity(marginal_density, occupancy, mean_rates[-1])
            + np.finfo(np.float32).eps)

        encoding_marks.append(
            multiunit[np.ix_(is_spike & not_nan_position, ~nan_mark_dims)
                      ].astype(np.float32))
        encoding_positions.append(position[is_spike & not_nan_position])

    summed_ground_process_intensity = np.sum(
        np.stack(ground_process_intensities, axis=0), axis=0, keepdims=True)

    return {
        'encoding_marks': encoding_marks,
        'encoding_positions': encoding_positions,
        'summed_ground_process_intensity': summed_ground_process_intensity,
        'occupancy': occupancy,
        'mean_rates': mean_rates,
        'mark_std': mark_std,
        'position_std': position_std,
        **kwargs,
    }



import cupy as cp
import numpy as np
from replay_trajectory_classification.bins import atleast_2d

SQRT_2PI = cp.sqrt(2.0 * cp.pi)


def gaussian_pdf(x, mean, sigma):
    '''Compute the value of a Gaussian probability density function at x with
    given mean and sigma.'''
    return cp.exp(-0.5 * ((x - mean) / sigma)**2) / (sigma * SQRT_2PI)


def estimate_position_distance(place_bin_centers, positions, position_std):
    return cp.prod(
        gaussian_pdf(
            cp.expand_dims(place_bin_centers, axis=0),
            cp.expand_dims(positions, axis=1),
            position_std),
        axis=-1
    )


def estimate_position_density(place_bin_centers, positions, position_std):
    return cp.mean(estimate_position_distance(
        place_bin_centers, positions, position_std), axis=0)


def estimate_log_intensity(density, occupancy, mean_rate):
    return cp.log(mean_rate) + cp.log(density) - cp.log(occupancy)


def estimate_intensity(density, occupancy, mean_rate):
    '''

    Parameters
    ----------
    density : ndarray, shape (n_bins,)
    occupancy : ndarray, shape (n_bins,)
    mean_rate : float

    Returns
    -------
    intensity : ndarray, shape (n_bins,)

    '''
    return cp.exp(estimate_log_intensity(density, occupancy, mean_rate))


def normal_pdf_integer_lookup(x, mean, std=20, max_value=6000, dtype=cp.float32):
    """Fast density evaluation for integers by precomputing a hash table of
    values.

    Parameters
    ----------
    x : int
    mean : int
    std : float
    max_value : int

    Returns
    -------
    probability_density : int

    """
    normal_density = gaussian_pdf(
        cp.arange(-max_value, max_value), 0, std).astype(dtype)

    return normal_density[(x - mean) + max_value]


def estimate_log_joint_mark_intensity(decoding_marks,
                                      encoding_marks,
                                      mark_std,
                                      occupancy,
                                      mean_rate,
                                      place_bin_centers=None,
                                      encoding_positions=None,
                                      position_std=None,
                                      max_mark_value=6000,
                                      set_diag_zero=False,
                                      position_distance=None):
    """

    Parameters
    ----------
    decoding_marks : ndarray, shape (n_decoding_spikes, n_features)
    encoding_marks : ndarray, shape (n_encoding_spikes, n_features)
    mark_std : float or ndarray, shape (n_features,)
    occupancy : ndarray, shape (n_position_bins,)
    mean_rate : float
    place_bin_centers : ndarray, shape (n_position_bins, n_position_dims)
    encoding_positions : ndarray, shape (n_decoding_spikes, n_position_dims)
    position_std : float or ndarray, shape (n_position_dims,)
    is_track_interior : None or ndarray, shape (n_position_bins,)
    max_mark_value : int
    set_diag_zero : bool

    Returns
    -------
    log_joint_mark_intensity : ndarray, shape (n_decoding_spikes, n_position_bins)

    """
    n_encoding_spikes, n_marks = encoding_marks.shape
    n_decoding_spikes = decoding_marks.shape[0]
    mark_distance = cp.ones((n_decoding_spikes, n_encoding_spikes))

    for mark_ind in range(n_marks):
        mark_distance *= normal_pdf_integer_lookup(
            cp.expand_dims(decoding_marks[:, mark_ind], axis=1),
            cp.expand_dims(encoding_marks[:, mark_ind], axis=0),
            std=mark_std,
            max_value=max_mark_value
        )

    if set_diag_zero:
        diag_ind = cp.diag_indices_from(mark_distance)
        mark_distance[diag_ind] = 0.0

    if position_distance is None:
        position_distance = estimate_position_distance(
            place_bin_centers, encoding_positions, position_std).astype(cp.float32)

    return cp.asnumpy(
        estimate_log_intensity(
            mark_distance @ position_distance / n_encoding_spikes,
            occupancy,
            mean_rate))


def fit_multiunit_likelihood_integer_cupy(position,
                                          multiunits,
                                          place_bin_centers,
                                          mark_std,
                                          position_std,
                                          is_track_interior=None,
                                          **kwargs):
    '''

    Parameters
    ----------
    position : ndarray, shape (n_time, n_position_dims)
    multiunits : ndarray, shape (n_time, n_marks, n_electrodes)
    place_bin_centers : ndarray, shape ( n_bins, n_position_dims)
    model : sklearn model
    model_kwargs : dict
    occupancy_model : sklearn model
    occupancy_kwargs : dict
    is_track_interior : None or ndarray, shape (n_bins,)

    Returns
    -------
    joint_pdf_models : list of sklearn models, shape (n_electrodes,)
    ground_process_intensities : list of ndarray, shape (n_electrodes,)
    occupancy : ndarray, (n_bins, n_position_dims)
    mean_rates : ndarray, (n_electrodes,)

    '''
    if is_track_interior is None:
        is_track_interior = np.ones((place_bin_centers.shape[0],),
                                    dtype=np.bool)
    position = atleast_2d(position)
    place_bin_centers = atleast_2d(place_bin_centers)

    not_nan_position = np.all(~np.isnan(position), axis=1)

    occupancy = np.zeros((place_bin_centers.shape[0],))
    occupancy[is_track_interior] = cp.asnumpy(
        estimate_position_density(
            cp.asarray(place_bin_centers[is_track_interior]),
            cp.asarray(position[not_nan_position]),
            position_std))

    mean_rates = []
    ground_process_intensities = []
    encoding_marks = []
    encoding_positions = []

    for multiunit in np.moveaxis(multiunits, -1, 0):

        # ground process intensity
        is_spike = np.any(~np.isnan(multiunit), axis=1)
        mean_rates.append(is_spike.mean())
        marginal_density = cp.zeros((place_bin_centers.shape[0],))

        if is_spike.sum() > 0:
            marginal_density[cp.asarray(is_track_interior)] = estimate_position_density(
                cp.asarray(place_bin_centers[is_track_interior]),
                cp.asarray(position[is_spike & not_nan_position]),
                position_std)

        ground_process_intensities.append(
            estimate_intensity(marginal_density, cp.asarray(
                occupancy), mean_rates[-1])
            + cp.asarray(np.spacing(1)))

        encoding_marks.append(
            cp.asarray(multiunit[is_spike & not_nan_position], dtype=cp.int16))
        encoding_positions.append(cp.asarray(
            position[is_spike & not_nan_position]))

    summed_ground_process_intensity = cp.asnumpy(cp.sum(
        cp.stack(ground_process_intensities, axis=0), axis=0, keepdims=True))

    return {
        'encoding_marks': encoding_marks,
        'encoding_positions': encoding_positions,
        'summed_ground_process_intensity': summed_ground_process_intensity,
        'occupancy': occupancy,
        'mean_rates': mean_rates,
        'mark_std': mark_std,
        'position_std': position_std,
        **kwargs,
    }


def estimate_multiunit_likelihood_integer_cupy(multiunits,
                                               encoding_marks,
                                               mark_std,
                                               place_bin_centers,
                                               encoding_positions,
                                               position_std,
                                               occupancy,
                                               mean_rates,
                                               summed_ground_process_intensity,
                                               max_mark_value=6000,
                                               set_diag_zero=False,
                                               is_track_interior=None,
                                               time_bin_size=1,
                                               block_size=None):
    '''

    Parameters
    ----------
    multiunits : ndarray, shape (n_time, n_marks, n_electrodes)
    place_bin_centers : ndarray, (n_bins, n_position_dims)
    joint_pdf_models : list of sklearn models, shape (n_electrodes,)
    ground_process_intensities : list of ndarray, shape (n_electrodes,)
    occupancy : ndarray, (n_bins, n_position_dims)
    mean_rates : ndarray, (n_electrodes,)

    Returns
    -------
    log_likelihood : (n_time, n_bins)

    '''

    if is_track_interior is None:
        is_track_interior = np.ones((place_bin_centers.shape[0],),
                                    dtype=np.bool)

    n_time = multiunits.shape[0]
    log_likelihood = (-time_bin_size * summed_ground_process_intensity *
                      np.ones((n_time, 1)))

    multiunits = np.moveaxis(multiunits, -1, 0)
    n_position_bins = is_track_interior.sum()

    for multiunit, enc_marks, enc_pos, mean_rate in zip(
            multiunits, encoding_marks, encoding_positions, mean_rates):
        is_spike = np.any(~np.isnan(multiunit), axis=1)
        decoding_marks = cp.asarray(multiunit[is_spike], dtype=cp.int16)
        n_decoding_marks = decoding_marks.shape[0]
        log_joint_mark_intensity = np.zeros(
            (n_decoding_marks, n_position_bins))

        if block_size is None:
            gpu_memory_size = cp.get_default_memory_pool().total_bytes()
            mark_distance_matrix_size = (
                n_decoding_marks * enc_marks.shape[0]) * 32 // 8  # bytes
            if mark_distance_matrix_size < gpu_memory_size * 0.8:
                if n_decoding_marks > 0:
                    block_size = n_decoding_marks
                else:
                    block_size = 1
            else:
                block_size = int(gpu_memory_size * 0.8 *
                                 8) // (32 * enc_marks.shape[0])

        position_distance = estimate_position_distance(
            cp.asarray(place_bin_centers[is_track_interior]),
            enc_pos,
            position_std).astype(cp.float32)

        for start_ind in range(0, n_decoding_marks, block_size):
            block_inds = slice(start_ind, start_ind + block_size)
            log_joint_mark_intensity[block_inds] = estimate_log_joint_mark_intensity(
                decoding_marks[block_inds],
                enc_marks,
                mark_std,
                cp.asarray(occupancy[is_track_interior]),
                mean_rate,
                max_mark_value=max_mark_value,
                set_diag_zero=set_diag_zero,
                position_distance=position_distance,
            )
        log_likelihood[np.ix_(is_spike, is_track_interior)] += (
            log_joint_mark_intensity + np.spacing(1))

    log_likelihood[:, ~is_track_interior] = np.nan

    return log_likelihood
